"""The main functionality of `image_to_excel`."""

import logging
from pathlib import Path

from .core.excel import ExcelWriter
from .core.image import ImageManipulator

logger = logging.getLogger(__name__)


class BaseClass:
    """Everything in the project comes back to here."""

    ACCEPTED_IMAGE_EXTENSIONS: list[str] = [".png", ".jpg", ".jpeg"]
    """The file extensions to accept for conversion."""

    def __init__(self, log_level: int = logging.INFO):
        """Initialises the base class for `image_to_excel` by loading the config and setting up a logger.

        Args:
            log_level (str): The level to use for package logs.
        """
        logging.getLogger("image_to_excel").setLevel(log_level)

    def validate_input_file(self, image_path: Path) -> None:
        """Validate the input file.

        Args:
            image_path (Path): The path to validate
        """
        if not image_path.exists():
            raise FileNotFoundError(f"Image file '{image_path}' does not exist.")

        if image_path.is_dir():
            raise ValueError(f"Image file '{image_path}' is a directory.")

        if image_path.suffix not in self.ACCEPTED_IMAGE_EXTENSIONS:
            raise ValueError(
                f"Image file '{image_path}' is not a valid image file.\n"
                f"Accepted extensions are: {', '.join(self.ACCEPTED_IMAGE_EXTENSIONS)}"
            )

    def validate_output_file(self, output_path: Path) -> None:
        """Validate the output file.

        Args:
            output_path (Path): The path to validate.
        """
        if not output_path.parent.exists():
            raise FileNotFoundError(f"Output directory '{output_path.parent}' does not exist.")

        if output_path.is_dir():
            raise ValueError(f"Output file '{output_path}' is a directory.")

        if output_path.suffix != ".xlsx":
            raise ValueError(f"Output file '{output_path}' is not a valid excel file. Please specify a '.xlsx' file.")

    def image_to_excel(self, image_path: Path, image_width: int, output_file_path: Path) -> None:
        """Convert an image to an excel file.

        Args:
            image_path (Path): The path to the image file to convert.
            image_width (int): The new width in pixels for the image.
                The height will be calculated automatically to maintain aspect ration.
            output_file_path (Path): The output excel file path.
        """
        image_manipulator = ImageManipulator()
        excel_writer = ExcelWriter(output_file_path)

        logger.info("Resizing image...")

        image_array = image_manipulator.image_to_data(image_path)
        image_array = image_manipulator.resize_image_to_width(image_array, image_width)

        logger.info("Writing image to excel file...")

        try:
            excel_writer.write_image_array(image_array)
        finally:
            excel_writer.close()

        logger.info("Image converted!")



""".. include:: ../../README.md"""  # noqa

import logging

from ._helpers import ColoredFormatter as _ColoredFormatter
from ._helpers import SuccessLogger as _SuccessLogger
from .base import BaseClass
from .cli import cli_main

# set up logging for the package
logging.setLoggerClass(_SuccessLogger)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
console = logging.StreamHandler()
console.setFormatter(_ColoredFormatter("[%(name)s] (%(levelname)s): %(message)s"))
logger.addHandler(console)

__all__ = ["BaseClass", "cli_main"]



"""CLI functionality of `image_to_excel`."""

import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path

from . import BaseClass


def cli_main() -> None:
    """CLI entrypoint for `image_to_excel`. Uses `BaseClass`."""
    argparser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    argparser.add_argument("-v", "--verbose", help="Enable verbose logging", action="store_true")
    argparser.add_argument("image_path", help="The path to the image file to convert")
    argparser.add_argument("output_file_path", help="The output excel file path")
    argparser.add_argument("-w", "--width", help="The new width in pixels for the image", type=int, default=100)

    args = argparser.parse_args()
    verbose_logging: bool = args.verbose
    image_path = Path(args.image_path)
    output_file_path = Path(args.output_file_path)
    image_width = args.width

    if verbose_logging:
        level = logging.DEBUG
    else:
        level = logging.INFO

    app = BaseClass(log_level=level)

    # simple validation bits
    if image_path == output_file_path:
        raise ValueError("The input and output file paths must be different.")

    app.validate_input_file(image_path)
    app.validate_output_file(output_file_path)

    app.image_to_excel(image_path, image_width, output_file_path)



"""Image manipulation module."""
import logging
from pathlib import Path

import numpy as np
from PIL import Image

logger = logging.getLogger(__name__)


class ImageManipulator:
    """Image Manipulator. Can load + resize images."""

    def image_to_data(self, image_path: Path) -> np.ndarray:
        """Load an image file from disk and return it as a numpy array.

        Args:
            image_path (Path): The location of the image file on disk.
        """
        logger.debug("Loading image from %s", image_path)
        with open(image_path, "rb") as image_file:
            image = Image.open(image_file)
            return np.asarray(image)

    def resize_image(self, image: np.ndarray, new_size: tuple[int, int]) -> np.ndarray:
        """Resize an image to a new size.

        Args:
            image (np.ndarray): The image to resize.
            new_size (tuple[int, int]): The new size for the image.
        """
        logger.debug("Resizing image to %s", new_size)
        return np.asarray(Image.fromarray(image).resize(new_size, Image.ANTIALIAS))

    def resize_image_to_width(self, image: np.ndarray, new_width: int) -> np.ndarray:
        """Resize an image to a new width.

        Args:
            image (np.ndarray): The image to resize.
            new_width (int): The new width for the image.
        """
        logger.debug("Resizing image to width %s", new_width)
        current_width, current_height = image.shape[1], image.shape[0]

        if new_width >= current_width:
            logger.warning(
                "New width is greater than or the same as the current width. " "Using original image without resizing."
            )
            return image

        new_height = int(current_height * (new_width / current_width))
        return self.resize_image(image, (new_width, new_height))



"""Helper functions and classes to make life easier. Mainly for internal use within the package."""

import logging

from .nice_logger import ColoredFormatter, SuccessLogger

logger = logging.getLogger(__name__)

__all__ = ["SuccessLogger", "ColoredFormatter"]



"""Nice logging features, including a "success" level and colored logging in both the terminal - and HTML."""

import copy
import html
import logging
from typing import Any, Dict

# Config stuff

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

COLORS = {
    "WARNING": YELLOW,
    "INFO": WHITE,
    "DEBUG": BLUE,
    "CRITICAL": YELLOW,
    "ERROR": RED,
    "SUCCESS": GREEN,
}

COLORS_HTML = {
    "WARNING": "yellow",
    "INFO": "black",
    "DEBUG": "grey",
    "CRITICAL": "darkred",
    "ERROR": "red",
    "SUCCESS": "green",
}

COLORS_BOOTSTRAP = {
    "WARNING": "warning",
    "INFO": "body",
    "DEBUG": "muted",
    "CRITICAL": "warning",
    "ERROR": "danger",
    "SUCCESS": "success",
}


RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

SUCCESS_LEVEL = 25


class ColoredFormatter(logging.Formatter):
    """A logging formatter which enables ANSI colors for each log level."""

    def format(self, record: logging.LogRecord) -> str:
        """Formats the `record`, ensuring coloured output.

        Args:
            record (logging.LogRecord): The log record to format

        Returns:
            str: The formatted log record
        """
        new_record = copy.copy(record)  # create a copy of the record so that the original is not changed

        levelname = new_record.levelname
        if levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            new_record.levelname = levelname_color
        return super().format(new_record)


class WebFormatter(logging.Formatter):
    """A logging formatter which enables ANSI colors for each log level - in HTML.

    Useful for situations where you want to show logging output in a web interface, or something like that.
    """

    COLORS: Dict[str, str] = {}
    """The find/replace map for log level colors"""

    def format(self, record: logging.LogRecord) -> str:
        """Formats the `record`, ensuring coloured output.

        Args:
            record (logging.LogRecord): The log record to format

        Returns:
            str: The formatted log record
        """
        new_record = copy.copy(record)  # create a copy of the record so that the original is not changed

        levelname = new_record.levelname
        if levelname in self.COLORS:
            replacement_fields = ["msg", "levelname"]

            for replacement_field in replacement_fields:
                value = new_record.__dict__[replacement_field]
                new_record.__dict__[replacement_field] = (
                    f"<span class='text-{self.COLORS[levelname]}'>" f"{html.escape(value)}" "</span>"
                )

        return super().format(new_record)


class HTMLFormatter(WebFormatter):
    """A logging formatter which enables ANSI colors for each log level - in HTML."""

    COLORS = COLORS_HTML


class BootstrapFormatter(WebFormatter):
    """A logging formatter which enables ANSI colors for each log level - in HTML (but with Bootstrap color names)."""

    COLORS = COLORS_BOOTSTRAP


class SuccessLogger(logging.Logger):
    """A logger which provides a `success` level - good to use with a `ColoredFormatter`."""

    def __init__(self, name: str, **kwargs: Any) -> None:
        """Initialises a `SuccessLogger`.

        Args:
            name (str): The name of the logger
        """
        super().__init__(name, **kwargs)

        # self.propagate = True  # don't propagate to parent loggers, otherwise we get duplicate messages

        logging.addLevelName(SUCCESS_LEVEL, "SUCCESS")

    def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
        """Enables the "success" log type for all loggers using `SuccessLogger`.

        Args:
            msg (str): The message to log
        """
        if self.isEnabledFor(SUCCESS_LEVEL):
            self._log(SUCCESS_LEVEL, msg, args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['image_to_excel', 'image_to_excel._helpers', 'image_to_excel.core']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.5,<2.0.0',
 'pillow>=9.3.0,<10.0.0',
 'pyyaml>=6.0,<7.0',
 'typeguard>=2.13.3,<3.0.0',
 'xlsxwriter>=3.0.3,<4.0.0']

entry_points = \
{'console_scripts': ['image-to-excel = image_to_excel:cli_main']}

setup_kwargs = {
    'name': 'image-to-excel',
    'version': '1.0.3',
    'description': 'A simple project to convert from an image to an excel file, because my brother wanted something that did this.',
    'long_description': '# image-to-excel\n\nA simple project to convert from an image to an excel file, because my brother wanted something that did this.\n\n## Installation\n\nYou can install the project using pip :)\n\n```bash\npip install image-to-excel\n```\n\n## Usage\n\nYou can use image-to-excel as an importable module:\n\n```py\nfrom image_to_excel import BaseClass\nfrom pathlib import Path\n\napp = BaseClass("config.yml")\n\napp.image_to_excel(\n    Path("input.jpg"), 100, Path("output.xlsx")\n)\n```\n\nOr as a command line interface:\n\n```bash\n$ python3 -m image_to_excel\n# or\n$ image-to-excel -w 100 input.jpg output.xslx\n```\n\n## Documentation\n\nDocumentation for image-to-excel can be found within the docs folder.\n',
    'author': 'alex.fayers',
    'author_email': 'alex@fayers.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/alexfayers/image-to-excel',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Module to write out data to a new excel file."""
import logging
from pathlib import Path

import numpy as np
import xlsxwriter
import xlsxwriter.format

logger = logging.getLogger(__name__)


class ExcelWriter:
    """Writes data to an excel file."""

    def __init__(self, output_path: Path) -> None:
        """Initialise the ExcelWriter.

        Args:
            output_path (Path): The path to write the excel file to.
        """
        logger.debug("Initialising ExcelWriter with output path %s", output_path)
        self.workbook = xlsxwriter.Workbook(output_path)
        self.worksheet = self.workbook.add_worksheet()

    def _get_pixel_format(self, pixel: np.ndarray) -> xlsxwriter.format.Format:
        """Get the format for a pixel.

        Args:
            pixel (np.ndarray): The pixel to get the format for.

        Returns:
            xlsxwriter.format.Format: The format for the pixel.
        """
        return self.workbook.add_format({"bg_color": f"#{pixel[0]:02x}{pixel[1]:02x}{pixel[2]:02x}"})

    def write_image_array(self, image: np.ndarray) -> None:
        """Write an image array to the excel file, using one cell to represent one pixel.

        Uses color formatting to set the color of each cell.

        Args:
            image (np.ndarray): The image to write to the worksheet.
        """
        total_rows: int = image.shape[0]

        for row_index, row in enumerate(image):
            for column_index, pixel in enumerate(row):
                self.worksheet.write(row_index, column_index, "", self._get_pixel_format(pixel))
            logger.debug("Wrote row %s of %s", row_index + 1, total_rows)

    def close(self) -> None:
        """Close the excel file."""
        logger.debug("Closing excel file")
        self.workbook.close()
